from urllib.parse import urlparse

from .classes.StiDataResult import StiDataResult
from .enums import StiDatabaseType
from .events.StiConnectionEventArgs import StiConnectionEventArgs
from .StiDataAdapter import StiDataAdapter


class StiMongoDbAdapter(StiDataAdapter):

### Properties

    version = '2025.1.4'
    checkVersion = True
    type = StiDatabaseType.MONGODB
    driverName = 'pymongo'


### Methods

    def connect(self):
        try:
            args = StiConnectionEventArgs(self.type, self.driverName, self.connectionInfo)
            self.handler.onDatabaseConnect(args)

            if args.link != None:
                self.connectionLink = args.link
            else:
                from pymongo import MongoClient
                self.connectionLink = MongoClient(self.connectionString)
        except Exception as e:
            return StiDataResult.getError(str(e)).getDataAdapterResult(self)
        
        return StiDataResult.getSuccess().getDataAdapterResult(self)
    
    def process(self):
        if super().process():
            return True
        
        url = urlparse(self.connectionString)
        self.connectionInfo.host = url.hostname
        self.connectionInfo.port = url.port
        self.connectionInfo.userId = url.username
        self.connectionInfo.password = url.password
        self.connectionInfo.database = url.path.strip(' /')

        parameterNames = []
        return self.processParameters(parameterNames)
    
    def processParameters(self, parameterNames: dict[str, list[str]]):
        return True
    
    def getType(self, meta: object):
        types = {
            'boolean': ['bool'],
            'int': ['int', 'long', 'minKey', 'maxKey'],
            'number': ['double', 'decimal'],
            'datetime': ['date', 'timestamp'],
            'array': ['binData', 'null'],
            'string': ['string', 'objectId', 'regex', 'javascript', 'array', 'object']
        }

        type = str(meta)
        for key, array in types.items():
            if type in array:
                return key

        return 'string'
    
    def getValue(self, value: object, valueType: str):
        if value is None:
            return None

        if valueType == 'string':
            return value if type(value) == str else str(value)

        return super().getValue(value, valueType)
    
    def executeNative(self, queryString: str, maxDataRows: int, result: StiDataResult):
        if not queryString:
            return self.retrieveSchema(result)

        return self.retrieveData(result, queryString, maxDataRows)
    
    def retrieveSchema(self, result: StiDataResult):
        from pymongo import MongoClient
        connectionLink: MongoClient = self.connectionLink

        database = connectionLink.get_database(self.connectionInfo.database)
        collectionNames = database.list_collection_names()

        schema: dict[str, dict] = {}
        for name in collectionNames:
            pipeline = [
                {'$project': {'_id': 0}},
                {'$project': {'data': {'$objectToArray': '$$ROOT'}}},
                {'$unwind': '$data'},
                {'$group': {
                    '_id': None,
                    'data': {'$addToSet': {'k': '$data.k', 'v': {'$type': '$data.v'}}}}
                },
                {'$replaceRoot': {'newRoot': {'$arrayToObject': '$data'}}}
            ]
            collection = database.get_collection(name)
            cursor = collection.aggregate(pipeline)
            for document in cursor:
                schema[name] = document
                break

        result.count = len(schema)
        for tableName, table in schema.items():
            for columnName, columnType in table.items():
                row = []
                row.append(tableName)
                row.append(columnName)
                row.append(self.getType(columnType))
                result.rows.append(row)

        return result

    def retrieveData(self, result: StiDataResult, queryString: str, maxDataRows: int):
        from pymongo import MongoClient
        connectionLink: MongoClient = self.connectionLink

        result = self.retrieveSchema(result)
        for item in result.rows:
            if item[0] == queryString:
                result.columns.append(item[1])
                result.types.append(item[2])

        result.count = len(result.columns)
        result.rows = []

        database = connectionLink.get_database(self.connectionInfo.database)
        collection = database.get_collection(queryString)
        cursor = collection.find()
        if maxDataRows != None:
            cursor.limit(maxDataRows)
        for document in cursor:
            row = [None] * result.count
            for key, value in document.items():
                try:
                    index = result.columns.index(key)
                    valueType = self.getValueType(value, index, result.types)
                    row[index] = self.getValue(value, valueType)
                except ValueError:
                    pass
            result.rows.append(row)
        cursor.close()

        result.count = len(result.rows)
        return result


from enum import Enum


class StiDatabaseType(Enum):

    NONE = None
    MYSQL = 'MySQL'
    MSSQL = 'MS SQL'
    POSTGRESQL = 'PostgreSQL'
    FIREBIRD = 'Firebird'
    ORACLE = 'Oracle'
    ODBC = 'ODBC'
    MONGODB = 'MongoDB'


### Helpers

    @staticmethod
    def getValues():
        return [enum.value for enum in StiDatabaseType if enum.value != None]


from enum import Enum, Flag


class StiEventArgs:

### Properties

    event: str = None
    """Name of the current event."""
    
    sender: object = None
    """The component that sent the request."""


### Helpers

    def __getProperties(self) -> list:
        return [name for name in dir(self) if not name.startswith('_')  and name != 'sender' and not callable(getattr(self, name))]

    def __setObject(self, object: object):
        properties = self.__getProperties()
        for name in properties:
            if hasattr(object, name):
                value = getattr(object, name)
                self.__setProperty(name, value)

    def __setProperty(self, name, value):
        selfvalue = getattr(self, name)
        if isinstance(selfvalue, Enum) or isinstance(value, Flag):
            if value != None: setattr(self, name, selfvalue.__class__(value))
        else: setattr(self, name, value)


### Constructor

    def __init__(self, object = None):
        if (object != None):
            self.__setObject(object)


class StiParameter:

### Properties

    name: str = None
    """The name of the parameter."""

    typeCode: int = None
    """The type code of the parameter."""

    typeName: str = None
    """The type name of the parameter."""

    typeGroup: str = None
    """The type group of the parameter."""

    size: int = None
    """The size of the parameter."""

    value: object = None
    """The value of the parameter. The type of object depends on the type of parameter."""


### Constructor

    def __init__(self, name: str, typeCode: int, typeName: str, typeGroup: str, size: int, value: object):
        self.name = name
        self.typeCode = typeCode
        self.typeName = typeName
        self.typeGroup = typeGroup
        self.size = size
        self.value = value
        


import codecs
import json
from enum import Enum, Flag

from ..enums import StiDatabaseType, StiDataCommand, StiBaseEventType
from .StiBaseResult import StiBaseResult


class StiBaseRequest:
    """Contains all set request parameters passed to the event handler."""

### Properties

    event = StiBaseEventType.NONE
    command = StiDataCommand.NONE
    encryptData = False
    connectionString: str = None
    queryString: str = None
    parameters: dict[str, object] = None
    database = StiDatabaseType.NONE
    dataSource: str = None
    connection: str = None
    timeout = 0
    maxDataRows: int = None
    escapeQueryParameters = False
    error: str = None


### Helpers

    def __getProperties(self) -> list:
        return [name for name in dir(self) if not name.startswith('_') and not callable(getattr(self, name))]

    def __setObject(self, object: dict, prefix = False):
        properties = self.__getProperties()
        for property in object:
            name = property[4:] if prefix and property[0:4] == 'sti_' else property
            if name in properties:
                self._setProperty(name, object.get(property))
    
    def _setProperty(self, name, value):
        selfvalue = getattr(self, name)
        if isinstance(selfvalue, Enum) or isinstance(value, Flag): setattr(self, name, selfvalue.__class__(value))
        else: setattr(self, name, value)


### Process

    def process(self, query: dict, body: str) -> bool:
        if len(query or '') > 0:
            self.__setObject(query, True)

        if len(body or '') > 0:
            if body[0] != '{':
                try:
                    body = codecs.decode(body, 'rot_13')
                    body = codecs.decode(body.encode(), 'base64').decode()
                except Exception as e:
                    self.error = 'Base64: ' + str(e)
                    return False
                
                self.encryptData = True

            try:
                obj = json.loads(body)
            except Exception as e:
                self.error = 'JSON: ' + str(e)
                return False

            self.__setObject(obj)

        return True

    def getResult(self):
        """
        The result of executing an event handler request. 
        The result contains a collection of data, message about the result of the command execution, and other technical information.
        """

        if self.error != None:
            return StiBaseResult.getError(self.error)

        StiBaseResult.getSuccess()
    


from __future__ import annotations

import typing

from .StiEventArgs import StiEventArgs

if typing.TYPE_CHECKING:
    from ..classes.StiBaseHandler import StiBaseHandler


class StiEvent:

### Fields

    __name: str = None
    __callbacks: list = None
    __handler: StiBaseHandler = None


### Properties
    
    @property
    def name(self):
        return self.__name
    
    @property
    def callbacks(self):
        return self.__callbacks
    
    @property
    def handler(self):
        return self.__handler
    

### Helpers

    def _setArgs(self, *args, **keywargs) -> StiEventArgs:
        eventArgs = args[0] if len(args) > 0 else keywargs.get('args')
        if isinstance(eventArgs, StiEventArgs):
            eventArgs.event = self.name[2:]
            eventArgs.sender = self.handler
            return eventArgs
        
        return None
    
    def append(self, callback) -> StiEvent:
        self.callbacks.append(callback)
        return self

    def remove(self, callback) -> StiEvent:
        self.callbacks.remove(callback)
        return self
    
    def hasServerCallbacks(self) -> bool:
        for callback in self.callbacks:
            if (callable(callback)):
                return True
        
        return False
    
    def hasClientCallbacks(self) -> bool:
        for callback in self.callbacks:
            if (isinstance(callback, str)):
                return True
        
        return False

### Override

    def __len__(self) -> int:
        return len(self.callbacks or '')

    def __iadd__(self, callback) -> StiEvent:
        return self.append(callback)
    
    def __isub__(self, callback) -> StiEvent:
        return self.remove(callback)

    def __call__(self, *args, **keywargs) -> object:
        for callback in self.callbacks:
            if (callable(callback)):
                self._setArgs(*args, **keywargs)
                result = callback(*args, **keywargs)
                if result != None:
                    return result
        return None
    
    def __contains__(self, callback):
        return callback in self.callbacks
	

### Constructor

    def __init__(self, handler: StiBaseHandler, name: str):
        self.__callbacks = []
        self.__handler = handler
        self.__name = name


from enum import Enum


class StiBaseEventType(Enum):

    NONE = None
    BEGIN_PROCESS_DATA = 'BeginProcessData'
    END_PROCESS_DATA = 'EndProcessData'


### Helpers

    @staticmethod
    def getValues(none = False):
        return [enum.value for enum in StiBaseEventType if none or enum.value != None]



from .classes.StiDataResult import StiDataResult
from .enums import StiDatabaseType
from .events.StiConnectionEventArgs import StiConnectionEventArgs
from .StiDataAdapter import StiDataAdapter


class StiOracleAdapter(StiDataAdapter):

### Properties

    version = '2025.1.4'
    checkVersion = True
    type = StiDatabaseType.ORACLE
    driverName = 'oracledb'


### Methods

    def getOdbcConnectionString(self):
        connectionString: str = \
            f'Driver={self.connectionInfo.driver};' \
            f'Dbq={self.connectionInfo.database};' \
            f'Uid={self.connectionInfo.userId};' \
            f'Pwd={self.connectionInfo.password};'
        
        return connectionString

    def connect(self):
        if self.connectionInfo.driver:
            return self.connectOdbc()
        
        try:
            args = StiConnectionEventArgs(self.type, self.driverName, self.connectionInfo)
            self.handler.onDatabaseConnect(args)
            
            if args.link != None:
                self.connectionLink = args.link
            else:
                import oracledb
                oracledb.init_oracle_client()
                self.connectionLink = oracledb.connect(
                    user = self.connectionInfo.userId,
                    password = self.connectionInfo.password,
                    dsn = self.connectionInfo.database,
                    encoding = self.connectionInfo.charset,
                    mode = self.connectionInfo.privilege)
        except Exception as e:
            return StiDataResult.getError(str(e)).getDataAdapterResult(self)
        
        return StiDataResult.getSuccess().getDataAdapterResult(self)
    
    def process(self):
        if super().process():
            return True
        
        import oracledb
        
        self.connectionInfo.port = 3306
        self.connectionInfo.privilege = oracledb.AUTH_MODE_DEFAULT

        parameterNames = {
            'driver': ['driver'],
            'database': ['database', 'data source', 'dbname'],
            'userId': ['uid', 'user', 'user id'],
            'password': ['pwd', 'password'],
            'charset': ['charset']
        }

        return self.processParameters(parameterNames)
    
    def processUnknownParameter(self, parameter: str, name: str, value: str):
        import oracledb
        
        super().processUnknownParameter(parameter, name, value)

        if name.lower() == 'dba privilege' or name.lower() == 'privilege':
            value = value.lower()
            if value == 'sysoper' or value == 'oci_sysoper':
                self.connectionInfo.privilege = oracledb.AUTH_MODE_SYSOPER
            if value == 'sysdba' or value == 'oci_sysdba':
                self.connectionInfo.privilege = oracledb.AUTH_MODE_SYSDBA

    def getType(self, meta: tuple):
        if self.connectionInfo.driver:
            return super().getType(meta)
        
        import oracledb
        info: oracledb.FetchInfo = meta
        
        types = {
            'int': [oracledb.DB_TYPE_NUMBER],
            'number': [oracledb.DB_TYPE_BINARY_DOUBLE, oracledb.DB_TYPE_BINARY_FLOAT],
            'time': [oracledb.DB_TYPE_INTERVAL_DS],
            'datetime': [oracledb.DB_TYPE_DATE, oracledb.DB_TYPE_TIMESTAMP, oracledb.DB_TYPE_TIMESTAMP_LTZ, oracledb.DB_TYPE_TIMESTAMP_TZ],
            'blob': [oracledb.DB_TYPE_BFILE, oracledb.DB_TYPE_BLOB, oracledb.DB_TYPE_CLOB, oracledb.DB_TYPE_CURSOR, oracledb.DB_TYPE_NCLOB],
            'array': [oracledb.DB_TYPE_LONG_RAW, oracledb.DB_TYPE_OBJECT, oracledb.DB_TYPE_RAW],
            'string': [oracledb.DB_TYPE_CHAR, oracledb.DB_TYPE_JSON, oracledb.DB_TYPE_LONG, oracledb.DB_TYPE_NCHAR, oracledb.DB_TYPE_NVARCHAR,
                       oracledb.DB_TYPE_ROWID, oracledb.DB_TYPE_ROWID, oracledb.DB_TYPE_UROWID, oracledb.DB_TYPE_VARCHAR]
        }

        for key, array in types.items():
            if info.type in array:
                if key == 'int':
                    return 'number' if info.scale > 0 else 'int'
                if key == 'blob':
                    return 'array'
                return key

        return 'string'
    
    def getValue(self, value: object, valueType: str):
        import oracledb
        
        if value is None:
            return None
        
        if valueType == 'array':
            if type(value) == oracledb.LOB:
                value = value.read()

        return super().getValue(value, valueType)

    def makeQuery(self, procedure: str, parameters: list):
        paramsString = super().makeQuery(procedure, parameters)
        return f"SQLEXEC 'CALL {procedure} {paramsString}'"


from __future__ import annotations


class StiBaseResult:
    """
    The result of executing an event handler request. 
    The result contains a collection of data, message about the result of the command execution, and other technical information.
    """

### Properties
    
    handlerVersion: str = None
    checkVersion = True
    success = True
    notice: str = None


### Abstract

    types: list


### JSON

    @staticmethod
    def getProperties(object: object):
        return { name: getattr(object, name) for name in dir(object) if not name.startswith('_') and not callable(getattr(object, name)) }


### Result

    @staticmethod
    def getSuccess(notice: str = None) -> StiBaseResult:
        """Creates a successful result."""

        result = StiBaseResult()
        result.success = True
        result.notice = notice
        return result
    
    @staticmethod
    def getError(notice: str) -> StiBaseResult:
        """Creates an error result."""

        result = StiBaseResult()
        result.success = False
        result.notice = notice
        return result